/* elflib.h -- analyze ELF executable and shared library
   Copyright 2000 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
#ifndef _h_elflib_h  /*{*/
#define _h_elflib_h

#include <elf.h>

extern int elf_good_header(Elf32_Ehdr const *const ehdr);

extern Elf32_Ehdr * elf_mmap(
	char const *const elf_name,
	size_t *p_elf_size
);

extern Elf32_Phdr const *elf_find_phdr_type(
	int key,
	Elf32_Ehdr const *const ehdr
);

extern Elf32_Shdr const * elf_find_section(
	void const *const ptr,
	Elf32_Ehdr const *const ehdr
);

extern Elf32_Shdr const * elf_find_section_name(
	char const *const name,
	Elf32_Ehdr const *const ehdr
);

extern Elf32_Shdr const * elf_find_section_type(
	int key,
	Elf32_Ehdr const *const ehdr
);

extern unsigned elf_get_dynamic(
	int const key,
	Elf32_Dyn const *dynp
);

extern unsigned elf_set_dynamic(
	int const key,
	unsigned const value,
	Elf32_Dyn *dynp
);

extern Elf32_Dyn const *elf_locate_dynamic(
	int const key,
	Elf32_Dyn const *dynp
);

extern void const * elf_find_dynamic(
	int const key,
	Elf32_Dyn const *dynp,
	Elf32_Ehdr const *const ehdr
);

extern char * elf_find_offset(
	unsigned const offset,
	Elf32_Ehdr const *const ehdr
);

extern unsigned elf_get_offset(
	unsigned const offset,
	Elf32_Ehdr const *const ehdr
);

extern unsigned elf_find_value(
	Elf32_Sym const *const sp,
	Elf32_Ehdr const *const ehdr
);

extern unsigned int elf_hash_dynsym(char const *p);

extern void elf_insert_dynsym(
	char const *const name,
	unsigned const ndx,
	unsigned *const htab
);

extern void elf_rehash(
	unsigned       *const new_htab,
	unsigned const *const htab,
	Elf32_Sym const *const dynsym,
	char const *const dynstr
);

#endif  /*}*/
