/* rpathrm.c -- remove temporary directories from DT_RPATH
   Copyright 2001 BitWagon Software LLC.  All rights reserved.
   This file may be used, modified, and distributed under the terms
   of the GNU General Public License version 2.
*/
#include <elf.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PAGE_MASK (~0u<<12)

char const **
colon_list(char const *const str)
{
	int n = 0;
	char const *p;
	if (0==str) {
		return 0;
	}
	for (p = str; 0!=*p; ++p) {
		n += (':' == *p);
	}
	{
		char const **const result = (char const **)malloc(
			(2+ n)*sizeof(*result));
		char *q = (char *)(result[n = 0] = strdup(str));
		for (; 0!=*q; ++q) {
			if (':'==*q) {
				*q++ = '\0';
				result[++n] = q;
			}
		}
		result[1+ n] = 0;
		return result;
	}
}

void
subtract(char const **x, char const *const *y)  /* x -= y */
{
	char const *p;
	for (p= *y++; 0!=p; p= *y++) {
		char const **q, **r;
		for (r= q= x; 0!=*q; ++q) {
			if (0!=strcmp(*q, p)) {
				*r++ = *q;
			}
		}
		*r = 0;
	}
}


/* rpathrm a.elf added_rpaths  # IN PLACE (overwrites a.elf) */
/*  If 2 args (no added_paths) then entire DT_RPATH is removed. */
int
main(int argc, char const *const *const argv)
{
	int const fd = open(argv[1], O_RDWR, 0);
	Elf32_Ehdr const *const ehdr = (Elf32_Ehdr const *)mmap(0, 4096,
		PROT_READ, MAP_PRIVATE, fd, 0);
	int phnum = ehdr->e_phnum;
	unsigned reloc0 = ~0;
	unsigned ph_base = ehdr->e_phoff & PAGE_MASK;
	unsigned ph_frag = ehdr->e_phoff - ph_base;
	Elf32_Phdr const *phdr = (Elf32_Phdr const *)(ph_frag +
		(char *)mmap(0, ph_frag + phnum * sizeof(Elf32_Phdr),
		PROT_READ, MAP_PRIVATE,
		fd, ph_base ) );
	int j;
	for (j= phnum; --j>=0; ++phdr)
	if (PT_LOAD==phdr->p_type && ~0==reloc0) {
		reloc0 = phdr->p_vaddr - phdr->p_offset;
	}
	phdr -= phnum;
	for (j = phnum; --j>=0; ++phdr) if (PT_DYNAMIC==phdr->p_type) {
		unsigned const pt_base = phdr->p_offset & PAGE_MASK;
		unsigned const pt_frag = phdr->p_offset - pt_base;
		int szDyn = phdr->p_filesz;
		Elf32_Dyn *dp_rpath    = 0;
		Elf32_Dyn *dp_relent   = 0;
		Elf32_Dyn *dp_strsz    = 0;
		Elf32_Dyn *dp_strtab   = 0;
		Elf32_Dyn *dp = (Elf32_Dyn *)(pt_frag +
			(char *)mmap(0, pt_frag + szDyn,
			PROT_READ|PROT_WRITE, MAP_SHARED,
			fd, pt_base ) );
		for ( ; 0<=(szDyn -= sizeof(*dp)); ++dp) {
			if (DT_RELENT ==dp->d_tag
			||  DT_RELAENT==dp->d_tag ) {
				dp_relent = dp;
			} else
			if (DT_RPATH==dp->d_tag) {
				dp_rpath = dp;
			} else
			if (DT_STRTAB==dp->d_tag) {
				dp_strtab = dp;
			} else
			if (DT_STRSZ==dp->d_tag) {
				dp_strsz = dp;
			}
		}
		if (0!=dp_rpath) {
			char const *const *extra;
			char const *      *have;
			char const *rpathstr;
			char *strtab;
			unsigned const unloc = dp_strtab->d_un.d_val - reloc0;
			unsigned const str_base = unloc & PAGE_MASK;
			unsigned const str_frag = unloc - str_base;
				
			if (0==dp_strtab || 0==dp_strsz) {
				fprintf(stderr,"no DT_STRTAB or DT_STRSZ in %s\n",
					argv[1]);
				return 1;
			}
			strtab = (char *)(str_frag +
				(char *)mmap(0, str_frag + dp_strsz->d_un.d_val,
				PROT_READ|PROT_WRITE, MAP_SHARED,
				fd, str_base ) );
			rpathstr = dp_rpath->d_un.d_val + strtab;
			have = (char const **)colon_list(rpathstr);
			extra = colon_list(argv[2] ? argv[2] : rpathstr);
			subtract(have, (extra ? extra : have));
			if (0==have[0] && 0!=dp_relent) {
				*dp_rpath = *dp_relent;
				return 0;
			}
			else {
				char *q = dp_rpath->d_un.d_val + strtab;
				char const *p;
				for (p= *have++; 0!=p; p= *have++) {
					unsigned const len = 1+ strlen(p);
					strncpy(q, p, len);
					q += len;
					if (0!=*have) {
						q[-1]=':';
					}
				}
			}
		}
	}
	return 0;
}

